"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const multilines_1 = __importDefault(require("multilines"));
const os_1 = require("os");
const scalars_1 = require("../static/scalars");
const utils_1 = require("../static/utils");
/**
 * Generates input types for the `seed` function.
 *
 * @param dmmf
 */
function generateGeneratedSeedModelsType(dmmf) {
    const models = dmmf.datamodel.models;
    const enums = dmmf.datamodel.enums;
    const generatedGenerateSeedModelsType = multilines_1.default `
  | "*": { amount: number }
  | ${models.map(generateSeedModelType).join(os_1.EOL)}
  `;
    return generatedGenerateSeedModelsType;
    /* Helper functions */
    /**
     * Generates type definitions of a particular model.
     *
     * @param model
     */
    function generateSeedModelType(model) {
        const fields = model.fields;
        /**
         * Factory has to be required on models that include non default
         * scalar fields.
         */
        const hasSupportedScalars = fields.filter(isScalar).every(isSupportedScalar);
        /* prettier-ignore */
        const generatedSeedModelType = multilines_1.default `
    | ${model.name}${q(hasSupportedScalars)}: { 
    |   amount?: number, 
    |   factory${q(hasSupportedScalars)}: {
    |     ${utils_1.filterMap(fields, f => generateSeedModelFieldType(model, f)).join(os_1.EOL)} 
    |   }
    | }
    `;
        return generatedSeedModelType;
    }
    /**
     * Generates type definitions of a particular model field.
     *
     * @param field
     */
    function generateSeedModelFieldType(model, field) {
        switch (field.kind) {
            case 'enum': {
                /**
                 * The field should generate a function which results in one
                 * of the enum values.
                 */
                const { values } = enums.find(e => e.name === field.type);
                const union = values.map(val => `"${val}"`).join(` | `);
                return `${field.name}?: (${union}) | (() => ${union})`;
            }
            case 'object': {
                /**
                 * The field should return a min or max number on abmigious
                 * fields.
                 */
                const relationType = getRelationType(model, field);
                switch (relationType) {
                    case '1-to-many': {
                        return `${field.name}?: { min?: number, max?: number }`;
                    }
                    case 'many-to-many': {
                        return `${field.name}?: { min?: number, max?: number }`;
                    }
                    case '1-to-1': {
                        if (field.isRequired)
                            return null;
                        return `${field.name}?: { min?: 0 | 1, max?: 0 | 1 }`;
                    }
                    case 'many-to-1': {
                        if (field.isRequired)
                            return null;
                        return `${field.name}?: { min?: 0 | 1, max?: 0 | 1 }`;
                    }
                    default: {
                        const never = relationType;
                        throw new Error('Not possible!');
                    }
                }
            }
            case 'scalar': {
                const scalar = getTSTypeFromDMMFScalar(field.type);
                const supported = isSupportedScalar(field);
                if (field.isList) {
                    /* prettier-ignore */
                    return `${field.name}${q(supported)}: ${scalar}[] | (() => ${scalar}[])`;
                }
                /**
                 * The field should provide a function which results in a type
                 * of the scalar.
                 */
                return `${field.name}${q(supported)}: ${scalar} | (() => ${scalar})`;
            }
            default: {
                const never = field.kind;
                throw new Error(`Received unexpected field kind "${field.kind}"`);
            }
        }
    }
    /**
     * Converts a DMMF scalar to TypeScript scalar.
     *
     * @param type
     */
    function getTSTypeFromDMMFScalar(type) {
        switch (type) {
            case scalars_1.Scalar.string: {
                return 'string';
            }
            case scalars_1.Scalar.int: {
                return 'number';
            }
            case scalars_1.Scalar.float: {
                return 'number';
            }
            case scalars_1.Scalar.bool: {
                return 'boolean';
            }
            case scalars_1.Scalar.date: {
                return 'string';
            }
            /* istanbul ignore next */
            default: {
                /* Returns any for unsupported/custom scalars. */
                return 'any';
            }
        }
    }
    /**
     * Calculates the relation type.
     *
     * @param field
     */
    function getRelationType(model, field) {
        /**
         * Find the back relation.
         */
        const relationModel = models.find(m => m.name === field.type);
        // const relationField = withDefault<DMMF.Field>(
        //   {
        //     kind: 'relation',
        //     name: '',
        //     isRequired: false,
        //     isList: false,
        //     isId: false,
        //     type: field.type,
        //     isGenerated: false,
        //     isUnique: false,
        //     dbName: '',
        //   },
        //   relationModel.fields.find(f => f.type === model.name),
        // )
        const relationField = relationModel.fields.find(f => f.type === model.name);
        /* Relation type definitions */
        if (field.isList && relationField.isList) {
            /**
             * many-to-many (A)
             *
             * model A {
             *  bs: [B]
             * }
             * model B {
             *  as: [A]
             * }
             */
            return 'many-to-many';
        }
        else if (!field.isList && relationField.isList) {
            /**
             * many-to-1 (A)
             *
             * model A {
             *  b: B
             * }
             * model B {
             *  as: [A]
             * }
             */
            return 'many-to-1';
        }
        else if (field.isList && !relationField.isList) {
            /**
             * 1-to-many (A)
             *
             * model A {
             *  bs: b[]
             * }
             *
             * model B {
             *  a: A
             * }
             */
            return '1-to-many';
        }
        else {
            /**
             * 1-to-1 (A)
             *
             * model A {
             *  b: B
             * }
             * model B {
             *  a: A
             * }
             */
            return '1-to-1';
        }
    }
    /**
     * Creates an optional TS field modifier.
     * "q" as question mark.
     *
     * @param isOptional
     */
    function q(isOptional) {
        switch (isOptional) {
            case true: {
                return '?';
            }
            case false: {
                return '';
            }
        }
    }
}
exports.generateGeneratedSeedModelsType = generateGeneratedSeedModelsType;
/* Helper functions */
/**
 * Determines whether a field scalar type is supported by default or not.
 * @param field
 */
function isSupportedScalar(field) {
    return Object.values(scalars_1.Scalar).some(s => s === field.type);
}
exports.isSupportedScalar = isSupportedScalar;
/**
 * Determines whether a field is a scalar.
 * @param field
 */
function isScalar(field) {
    return field.kind === 'scalar';
}
exports.isScalar = isScalar;
//# sourceMappingURL=seed.js.map