"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = __importDefault(require("lodash"));
const mysql_1 = __importDefault(require("mysql"));
const url_1 = __importDefault(require("url"));
const lift_1 = require("../lift");
const pool_1 = require("../pool");
/**
 * Creates a dmmf specific Internal Pool instance.
 *
 * @param dmmf
 */
function getMySQLPool(dmmf, cwd) {
    return class extends MySQLPool {
        constructor(options) {
            super(dmmf, Object.assign({ tmpPrismaSchemaPath: lift_1.getTmpPrismaSchemaPath }, options), cwd);
        }
    };
}
exports.getMySQLPool = getMySQLPool;
class MySQLPool extends pool_1.InternalPool {
    constructor(dmmf, options, cwd) {
        super({ max: lodash_1.default.get(options, ['pool', 'max'], Infinity) });
        this.dmmf = dmmf;
        this.projectDir = cwd;
        this.getConnection = options.connection;
        this.getTmpPrismaSchemaPath = options.tmpPrismaSchemaPath;
    }
    createDBInstance(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const connection = this.getConnection(id);
            const tmpPrismaSchemaPath = this.getTmpPrismaSchemaPath(id);
            const uri = readMySQLURI(connection);
            const datasources = [
                {
                    name: 'db',
                    connectorType: 'mysql',
                    url: {
                        value: uri,
                        fromEnvVar: null,
                    },
                    config: {},
                },
            ];
            /* Migrate using Lift. */
            const { datamodel } = yield lift_1.migrateLift({
                id,
                projectDir: this.projectDir,
                datasources,
                tmpPrismaSchemaPath,
                dmmf: this.dmmf,
            });
            const instance = {
                url: uri,
                cwd: this.projectDir,
                datamodel: datamodel,
            };
            return instance;
        });
    }
    deleteDBInstance(instance) {
        return __awaiter(this, void 0, void 0, function* () {
            const connection = parseMySQLURI(instance.url);
            try {
                const client = yield getMySQLClient(connection);
                yield query(client, `DROP DATABASE IF EXISTS \`${connection.database}\`;`);
            }
            catch (err) /* istanbul ignore next */ {
                throw err;
            }
        });
    }
}
/* Helper functions */
/**
 * Creates a mysql.Connection instance and makes sure it's connected.
 *
 * @param connection
 */
function getMySQLClient(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        const client = mysql_1.default.createConnection({
            host: connection.host,
            port: connection.port,
            user: connection.user,
            password: connection.password,
            database: connection.database,
        });
        return new Promise((resolve, reject) => {
            client.connect(err => {
                /* istanbul ignore if */
                if (err)
                    reject(err);
                else
                    resolve(client);
            });
        });
    });
}
/**
 * Executes a query as a promise against MySQL db with established conneciton.
 *
 * @param connection
 * @param query
 */
function query(client, query) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            client.query(query, (err, res) => {
                /* istanbul ignore if */
                if (err)
                    reject(err);
                else
                    resolve(res);
            });
        });
    });
}
/**
 * Converts a MySQLConnection into a MySQL URI.
 *
 * @param connection
 */
function readMySQLURI(connection) {
    return `mysql://${connection.user}:${connection.password}@${connection.host}:${connection.port}/${connection.database}`;
}
/**
 * Parses MySQL URI into MySQLConnection.
 *
 * @param uri
 */
function parseMySQLURI(uri) {
    const { auth, hostname: host, port: rawPort, pathname } = url_1.default.parse(uri, true);
    const [, user, password] = auth.match(/(\w+):(\w+)/);
    const [, database] = pathname.match(/\/(.+)/);
    const port = parseInt(rawPort, 10);
    return {
        user,
        password,
        host: host,
        port,
        database,
    };
}
//# sourceMappingURL=mysql.js.map