"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
class InternalPool {
    constructor(options) {
        this.dbs = {
            booting: [],
            busy: [],
        };
        this.waiters = [];
        this.capacity = options.max;
    }
    /**
     * Creates a new DBInstance and occupies the space in the pool.
     */
    getDBInstance() {
        return __awaiter(this, void 0, void 0, function* () {
            /**
             * If full capacity is not yet reached create new instance.
             * Otherwise, return a waiter which will wait until there's
             * an available space.
             */
            if (this.dbs.busy.length + this.dbs.booting.length < this.capacity) {
                /* Generates unique DB identifier. */
                const id = Math.random()
                    .toString(36)
                    .slice(2);
                /* Creates a new DBInstance. */
                this.dbs.booting = this.dbs.booting.concat(id);
                const dbInstance = yield this.createDBInstance(id);
                this.dbs.booting = this.dbs.booting.filter(dbId => dbId !== id);
                this.dbs.busy = this.dbs.busy.concat(dbInstance);
                return dbInstance;
            }
            else {
                /* Add to the line. */
                const waiter = new Waiter();
                this.waiters = this.waiters.concat(waiter);
                return waiter.wait();
            }
        });
    }
    /**
     * Releases a db isntance in the pool and triggers the creation
     * of new instance if there are waiters.
     *
     * @param db
     */
    releaseDBInstance(instance) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                /* Finds the busy instance and releases it. */
                yield this.deleteDBInstance(instance);
                this.dbs.busy = this.dbs.busy.filter(db => db.url !== instance.url);
                /**
                 * Allocates a new db instance for a waiter.
                 */
                const [waiter, ...remainingWaiters] = this.waiters;
                if (waiter) {
                    const instance = yield this.getDBInstance();
                    waiter.allocate(instance);
                    this.waiters = remainingWaiters;
                }
            }
            catch (err) /* istanbul ignore next */ {
                throw err;
            }
        });
    }
    /**
     * A function wrapper which makes sure that the db instance is
     * correctly allocated prior to execution of the function, and correctly
     * released after the execution has finished.
     *
     * @param fn
     */
    run(fn) {
        return __awaiter(this, void 0, void 0, function* () {
            const db = yield this.getDBInstance();
            try {
                const result = yield fn(db);
                yield this.releaseDBInstance(db);
                return result;
            }
            catch (e) {
                yield this.releaseDBInstance(db);
                throw e;
            }
        });
    }
    /**
     * Rejects remaining waiters, and releases all
     * remaining instances in the pool.
     */
    drain() {
        return __awaiter(this, void 0, void 0, function* () {
            /* Reject waiters. */
            this.waiters.forEach(w => w.fail('Drained before allocated.'));
            this.waiters = [];
            /* Release busy instances. */
            const actions = this.dbs.busy.map(i => this.releaseDBInstance(i));
            try {
                yield Promise.all(actions);
            }
            catch (err) /* istanbul ignore next */ {
                throw err;
            }
        });
    }
}
exports.InternalPool = InternalPool;
/* Pools specific to a DB type. */
class Waiter {
    constructor() {
        this.promise = new Promise((resolve, reject) => {
            this.resolve = resolve;
            this.reject = reject;
        });
    }
    wait() {
        return this.promise;
    }
    allocate(instance) {
        this.resolve(instance);
    }
    fail(reason) {
        this.reject(reason);
    }
}
//# sourceMappingURL=pool.js.map