"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const chance_1 = __importDefault(require("chance"));
const lodash_1 = __importDefault(require("lodash"));
const multilines_1 = __importDefault(require("multilines"));
const scalars_1 = require("./scalars");
const utils_1 = require("./utils");
/**
 * Creates a function which can be used to seed mock data to database.
 *
 * @param dmmf
 */
function getSeed(dmmf) {
    return (options) => __awaiter(this, void 0, void 0, function* () {
        /**
         * The wrapped function which handles the execution of
         * the seeding algorithm.
         */
        const opts = Object.assign({ seed: 42 }, options);
        const faker = new chance_1.default(opts.seed);
        const kit = {
            faker,
        };
        const models = options.models
            ? options.models(kit)
            : {
                '*': {
                    amount: 5,
                },
            };
        /* Fixture calculations */
        const orders = getOrdersFromDMMF(dmmf, models);
        const steps = getStepsFromOrders(orders);
        const tasks = getTasksFromSteps(steps);
        /* Creates mock data and pushes it to Prisma */
        const fixtures = yield seedTasks(options.client, faker, models, orders, tasks);
        return groupFixtures(fixtures);
    });
    /* Helper functions */
    /**
     * Converts dmmf to orders. Orders represent the summary of a mock requirements
     * for particular model. During generation the function detects relation types,
     * and validates the bulk of the request.
     *
     * @param dmmf
     */
    function getOrdersFromDMMF(dmmf, seedModels) {
        return dmmf.datamodel.models.map(model => {
            /* User defined settings */
            const fakerModel = getSeedModel(seedModels, model.name);
            /* Find Photon mappings for seeding step */
            const mapping = dmmf.mappings.find(m => m.model === model.name);
            /* Generate relations based on provided restrictions. */
            const relations = model.fields
                .filter(f => f.kind === 'object')
                .reduce((acc, field) => {
                const seedModelField = lodash_1.default.get(fakerModel, ['factory', field.name], {});
                switch (typeof seedModelField) {
                    case 'object': {
                        /* Calculate the relation properties */
                        return Object.assign(Object.assign({}, acc), { [field.name]: getRelationType(dmmf.datamodel.models, seedModels, field, model, seedModelField) });
                    }
                    /* istanbul ignore next */
                    default: {
                        throw new Error(`Expected a relation constraint got ${typeof seedModelField}`);
                    }
                }
            }, {});
            const enums = model.fields
                .filter(field => field.kind === 'enum')
                .reduce((acc, field) => {
                const enume = dmmf.datamodel.enums.find(enume => enume.name === field.type);
                if (enume === undefined) {
                    throw new Error(`Couldn't find enumerator declaration for ${field.type}`);
                }
                return Object.assign(Object.assign({}, acc), { [field.name]: enume });
            }, {});
            return {
                model: model,
                mapping: mapping,
                amount: fakerModel.amount,
                relations: relations,
                enums: enums,
            };
        });
        /**
         * Finds a model definition in seed models or returns the required
         * fallback seed model definition.
         */
        function getSeedModel(seedModels, model) {
            const fallback = seedModels['*'];
            const definitionConstructor = lodash_1.default.get(seedModels, model);
            return utils_1.withDefault(fallback, Object.assign({ amount: seedModels['*'].amount }, definitionConstructor));
        }
        /**
         * Finds the prescribed model from the DMMF models.
         */
        function getDMMFModel(models, model) {
            return models.find(m => m.name === model);
        }
        /**
         * Derives the relation type, relation direction, and constraints from the field
         * and definition.
         *
         * We have four different relation types:
         * 1. 1-to-1
         *  ~ There can be at most one connection or none if optional.
         *  ~ We'll connect the nodes from this model (a child model).
         * 2. 1-to-many
         *  ~ The node can have 0 to infinite connections (we'll use min/max spec).
         *  ~ We'll connect the nodes from this model (a child model).
         * 3. many-to-1
         *  ~ A particular instance can either connect or not connect if optional.
         *  ~ We'll fill the pool with ids of this model (a parent model) or create nodes
         *    if the relation is required in both directions.
         * 4. many-to-many
         *  ~ A particular instance can have 0 to infinite connections (we'll use min/max spec).
         *  ~ We'll use `relationTo` to determine whether this is a parent or child model and
         *    create nodes accordingly.
         *
         * We presume that the field is a relation.
         */
        function getRelationType(dmmfModels, seedModels, field, fieldModel, definition) {
            /**
             * model A {
             *  field: Relation
             * }
             */
            /* Field definitions */
            const fieldSeedModel = getSeedModel(seedModels, fieldModel.name);
            /**
             * Relation definitions
             *
             * NOTE: relaitonField is a back reference to the examined model.
             */
            const relationModel = getDMMFModel(dmmfModels, field.type);
            const backRelationField = relationModel.fields.find(f => f.relationName === field.relationName);
            const relationSeedModel = getSeedModel(seedModels, field.type);
            /* Relation type definitions */
            if (field.isList && backRelationField.isList) {
                /**
                 * many-to-many (A)
                 *
                 * model A {
                 *  bs: B[]
                 * }
                 * model B {
                 *  as: A[]
                 * }
                 */
                const min = utils_1.withDefault(0, definition.min);
                const max = utils_1.withDefault(min, definition.max);
                /* Validation */
                if (min > max) {
                    /* istanbul ignore next */ /* Inconsistent mock definition. */
                    throw new Error(
                    /* prettier-ignore */
                    multilines_1.default `
            | ${fieldModel.name}.${field.name}: number of minimum instances is higher than maximum.
            `);
                }
                else if (max > relationSeedModel.amount) {
                    /* istanbul ignore next */ /* Missing relation instances */
                    const missingInstances = max - relationSeedModel.amount;
                    throw new Error(
                    /* prettier-ignore */
                    multilines_1.default `
            | ${fieldModel.name}.${field.name} requests more(${max}) instances of | ${relationModel.name}(${relationSeedModel.amount}) than available.
            | Please add more(${missingInstances}) ${relationModel.name} instances.
            `);
                }
                else {
                    /* Valid declaration */
                    return {
                        type: 'many-to-many',
                        min: min,
                        max: max,
                        direction: getRelationDirection(field, backRelationField),
                        field,
                        backRelationField: backRelationField,
                    };
                }
            }
            else if (!field.isList && backRelationField.isList) {
                /**
                 * many-to-1 (A)
                 *
                 * model A {
                 *  b: B
                 * }
                 * model B {
                 *  as: [A]
                 * }
                 */
                if (field.isRequired &&
                    relationSeedModel.amount < fieldSeedModel.amount) {
                    const missingInstances = fieldSeedModel.amount - relationSeedModel.amount;
                    throw new Error(
                    /* prettier-ignore */
                    multilines_1.default `
            | ${fieldModel.name}.${field.name} requests more (${fieldSeedModel.amount}) instances of ${relationModel.name}(${relationSeedModel.amount}) than available.
            | Please add ${missingInstances} more ${relationModel.name} instances.
            `);
                }
                const min = field.isRequired ? 1 : utils_1.withDefault(0, definition.min);
                return {
                    type: 'many-to-1',
                    min: min,
                    max: 1,
                    direction: getRelationDirection(field, backRelationField),
                    field,
                    backRelationField: backRelationField,
                };
            }
            else if (field.isList && !backRelationField.isList) {
                /**
                 * 1-to-many (A)
                 *
                 * model A {
                 *  bs: b[]
                 * }
                 *
                 * model B {
                 *  a: A
                 * }
                 */
                /**
                 * TODO: This is not completely accurate, because there
                 * could be more of this type than of backRelation, and this
                 * setup doesn't allow for that.
                 *
                 * TODO: Intellisense: if back relation field required min should be at least 1.
                 */
                // const min = backRelationField.isRequired
                //   ? 1
                //   : withDefault(0, definition.min)
                const min = utils_1.withDefault(0, definition.min);
                const max = utils_1.withDefault(fieldSeedModel.amount, definition.max);
                /* Validation */
                if (min > max) {
                    /* istanbul ignore next */ /* Inconsistent mock definition. */
                    throw new Error(
                    /* prettier-ignore */
                    multilines_1.default `
            | ${fieldModel.name}.${field.name}: number of minimum instances is higher than maximum.
            `);
                }
                // else if (max > relationSeedModel.amount) {
                //   /* istanbul ignore next */ /* Missing relation instances */
                //   const missingInstances = max - relationSeedModel.amount
                //   throw new Error(
                //     /* prettier-ignore */
                //     mls`
                //     | ${fieldModel.name}.${field.name} requests more (${max}) instances of ${relationModel.name}(${relationSeedModel.amount}) than available.
                //     | Please add more (${missingInstances}) ${relationModel.name} instances.
                //     `,
                //   )
                // }
                if (min * fieldSeedModel.amount > relationSeedModel.amount) {
                    const missingInstances = min * fieldSeedModel.amount - relationSeedModel.amount;
                    throw new Error(
                    /* prettier-ignore */
                    multilines_1.default `
            | ${fieldModel.name}.${field.name} requests more (${min * fieldSeedModel.amount}) instances of ${relationModel.name}(${relationSeedModel.amount}) than available.
            | Please add ${missingInstances} more ${relationModel.name} instances.
            `);
                }
                /* Valid declaration */
                return {
                    type: '1-to-many',
                    min: min,
                    max: max,
                    direction: getRelationDirection(field, backRelationField),
                    field,
                    backRelationField: backRelationField,
                };
            }
            else {
                /**
                 * 1-to-1 (A)
                 *
                 * model A {
                 *  b: B
                 * }
                 * model B {
                 *  a: A
                 * }
                 */
                /* Validation */
                if (field.isRequired &&
                    backRelationField.isRequired &&
                    fieldSeedModel.amount !== relationSeedModel.amount) {
                    /* Required 1-to-1 relation unit amount mismatch. */
                    throw new Error(
                    /* prettier-ignore */
                    multilines_1.default `
            | A 1-to-1 required relation ${fieldModel.name}.${field.name}-${relationModel.name} has different number of units assigned.
            | Please make sure that number of ${fieldModel.name} and ${relationModel.name} match.
            `);
                }
                else if (!field.isRequired &&
                    backRelationField.isRequired &&
                    fieldSeedModel.amount < relationSeedModel.amount) {
                    /* An optional 1-to-1 relation inadequate unit amount. */
                    throw new Error(
                    /* prettier-ignore */
                    multilines_1.default `
            | A 1-to-1 relation ${relationModel.name} needs at least ${relationSeedModel.amount} ${fieldModel.name} units, but only ${fieldSeedModel.amount} were provided.
            | Please make sure there's an adequate amount of resources available.
            `);
                }
                else {
                    /* Sufficient amounts. */
                    return {
                        type: '1-to-1',
                        min: field.isRequired ? 1 : utils_1.withDefault(0, definition.min),
                        max: 1,
                        direction: getRelationDirection(field, backRelationField),
                        field,
                        backRelationField: backRelationField,
                    };
                }
            }
        }
        /**
         * Determines relation direction based on the type of fields
         * connecting the two types together.
         *
         * `relation direction` tells which of the two models should be created first.
         */
        function getRelationDirection(field, backRelationField) {
            /* Model of the observed type. */
            const fieldModel = backRelationField.type;
            const relationModel = field.type;
            if (field.isRequired && backRelationField.isRequired) {
                /**
                 * model A {
                 *  b: B
                 * }
                 * model B {
                 *  a: A
                 * }
                 *
                 * -> Create B while creating A, the order doesn't matter.
                 */
                return {
                    optional: true,
                    from: lodash_1.default.head([fieldModel, relationModel].sort()),
                };
            }
            else if (field.isRequired && !backRelationField.isRequired) {
                /**
                 * model A {
                 *  b: B
                 * }
                 * model B {
                 *  a: A?
                 * }
                 *
                 * We should create B first.
                 */
                return {
                    optional: false,
                    from: relationModel,
                };
            }
            else if (field.isRequired && backRelationField.isList) {
                /**
                 * model A {
                 *  b: B
                 * }
                 * model B {
                 *  a: A[]
                 * }
                 *
                 * -> We should create B and connect As to it.
                 */
                return {
                    optional: false,
                    from: relationModel,
                };
            }
            else if (field.isList && backRelationField.isRequired) {
                /**
                 * model A {
                 *  b: B[]
                 * }
                 * model B {
                 *  a: A
                 * }
                 *
                 * -> We should create A and connect Bs to it later.
                 */
                return {
                    optional: false,
                    from: fieldModel,
                };
            }
            else if (field.isList && !backRelationField.isRequired) {
                /**
                 * model A {
                 *  b: B[]
                 * }
                 * model B {
                 *  a: A?
                 * }
                 *
                 * -> We should create B first.
                 */
                return {
                    optional: true,
                    from: lodash_1.default.head([fieldModel, relationModel].sort()),
                };
            }
            else if (field.isList && backRelationField.isList) {
                /**
                 * model A {
                 *  b: B[]
                 * }
                 * model B {
                 *  a: A[]
                 * }
                 *
                 * -> The order doesn't matter, just be consistent.
                 */
                return {
                    optional: true,
                    from: lodash_1.default.head([fieldModel, relationModel].sort()),
                };
            }
            else if (!field.isRequired && backRelationField.isRequired) {
                /**
                 * model A {
                 *  b: B?
                 * }
                 * model B {
                 *  a: A
                 * }
                 *
                 * We should create A first, and connect B with A once we create B.
                 */
                return {
                    optional: false,
                    from: fieldModel,
                };
            }
            else if (!field.isRequired && !backRelationField.isRequired) {
                /**
                 * model A {
                 *  b: B?
                 * }
                 * model B {
                 *  a: A?
                 * }
                 *
                 * -> The order doesn't matter just be consistent.
                 */
                return {
                    optional: true,
                    from: lodash_1.default.head([fieldModel, relationModel].sort()),
                };
            }
            else if (!field.isRequired && backRelationField.isList) {
                /**
                 * model A {
                 *  b: B?
                 * }
                 * model B {
                 *  a: A[]
                 * }
                 *
                 * -> We should create A(s) first and then connect B with them.
                 */
                return {
                    optional: true,
                    from: lodash_1.default.head([fieldModel, relationModel].sort()),
                };
            }
            else {
                throw new Error(`Uncovered relation type.`);
            }
        }
    }
    /**
     * Coverts orders to steps. Steps represent an ordered entity that specifies the creation
     * of a virtual data.
     *
     * Creates a pool of available instances and validates relation constraints. Using
     * `isOrderWellDefined` and `getStepsFromOrder` functions, you should be able to implement
     * the topological sort algorithm.
     *
     * This function assumes that the graph is acylic.
     *
     * @param orders
     */
    function getStepsFromOrders(orders) {
        let graph = orders.filter(utils_1.not(isLeafOrder));
        let pool = {}; // edges
        let sortedSteps = []; // L
        let leafs = orders.filter(isLeafOrder); // S
        while (leafs.length !== 0) {
            const leaf = leafs.shift(); // n
            const { pool: poolWithLeaf, step: leafStep } = insertOrderIntoPool(sortedSteps.length, pool, leaf);
            pool = poolWithLeaf;
            sortedSteps.push(leafStep);
            for (const order of graph) {
                if (isOrderWellDefinedInPool(pool, order)) {
                    /* Remove the edge from the graph. */
                    graph = graph.filter(({ model }) => model.name !== order.model.name);
                    /* Update the leafs list. */
                    leafs.push(order);
                }
            }
        }
        if (graph.length !== 0) {
            throw new Error(`${graph.map(o => o.model.name).join(', ')} have cycles!`);
        }
        return sortedSteps;
        /* Helper functions */
        /**
         * Determines whether we can already process the order based on the pool
         * capacity.
         *
         * This function in combination with `getStepsFromOrder` should give you
         * all you need to implement meaningful topological sort on steps.
         */
        function isOrderWellDefinedInPool(pool, order) {
            return Object.values(order.relations).every(relation => pool.hasOwnProperty(relation.field.type) ||
                relation.direction.from === order.model.name ||
                relation.direction.optional);
        }
        /**
         * Tells whether the order has outgoing relations.
         *
         * @param order
         */
        function isLeafOrder(order) {
            return Object.values(order.relations).every(relation => relation.direction.from === order.model.name ||
                relation.direction.optional);
        }
        /**
         * Converts a well defined order to a step and adds it to the pool.
         *
         * This function in combination with `isOrderWellDefinedInPool` should give
         * you everything you need to implement meaningful topological sort on steps.
         */
        function insertOrderIntoPool(ordinal, pool, order) {
            /**
             * Fix optional relations' direction.
             */
            const fixedRelations = utils_1.mapEntries(order.relations, relation => {
                if (!relation.direction.optional)
                    return relation;
                return Object.assign(Object.assign({}, relation), { direction: {
                        optional: false,
                        /**
                         * If pool already includes the model of the relation,
                         * make the foreign relation first and create this model
                         * afterwards.
                         */
                        from: pool.hasOwnProperty(relation.field.type)
                            ? /* back relation model */ relation.field.type
                            : /* this model */ relation.backRelationField.type,
                    } });
            });
            /**
             * A step unit derived from the order.
             */
            const step = {
                order: ordinal,
                amount: order.amount,
                model: order.model,
                mapping: order.mapping,
                relations: fixedRelations,
                enums: order.enums,
            };
            /**
             * Assumes that there cannot exist two models with the same name.
             */
            const newPool = Object.assign(Object.assign({}, pool), { [order.model.name]: order.model });
            return {
                step,
                pool: newPool,
            };
        }
    }
    /**
     * Converts steps to tasks.
     *
     * Steps to Tasks introduce no non-trivial logic. It's a simple conversion mechanism
     * to make system more robuts and make the mocking part more granular.
     *
     * @param steps
     */
    function getTasksFromSteps(steps) {
        const tasks = steps.reduce((acc, step) => {
            const intermediateTasks = Array(step.amount).fill({
                order: step.order,
                model: step.model,
                mapping: step.mapping,
                relations: step.relations,
                enums: step.enums,
            });
            return acc.concat(...intermediateTasks);
        }, []);
        const tasksWithCorrectOrder = lodash_1.default.sortBy(tasks, t => t.order).map((task, i) => (Object.assign(Object.assign({}, task), { order: i })));
        return tasksWithCorrectOrder;
    }
    /**
     * Converts tasks to fixtures by creating a pool of available instances
     * and assigning relations to particular types.
     *
     * This function assumes that:
     *  1. Tasks (Steps) are sorted in such an order that pool always possesses
     *    all required instances,
     *  2. There are enough resources in the pool at all times,
     *  3. The provided schema is valid.
     *
     * @param steps
     */
    function seedTasks(client, faker, seedModels, orders, tasks) {
        return __awaiter(this, void 0, void 0, function* () {
            const initialPool = {};
            const { fixtures } = yield iterate(lodash_1.default.sortBy(tasks, t => t.order), initialPool);
            return fixtures;
            /**
             * Recursively seeds tasks to database.
             */
            function iterate([currentTask, ...remainingTasks], availablePool, iteration = 0) {
                return __awaiter(this, void 0, void 0, function* () {
                    /* Edge case */
                    if (currentTask === undefined) {
                        return {
                            fixtures: [],
                            pool: availablePool,
                        };
                    }
                    /* Recursive step */
                    /* Fixture calculation */
                    const { data, pool: newPool, tasks: newTasks, include } = getMockForTask(availablePool, remainingTasks, currentTask);
                    const methodName = lodash_1.default.lowerFirst(currentTask.mapping.model);
                    /**
                     * Make sure that client packs everyting.
                     */
                    if (!client[methodName]) {
                        throw new Error(`Client is missing method for ${currentTask.model.name} (methodName = ${methodName}, mappingModel = ${currentTask.mapping.model})`);
                    }
                    /**
                     * Load the data to database.
                     */
                    const seed = yield client[methodName].create(Object.assign({ data }, (Object.keys(include).length
                        ? {
                            include,
                        }
                        : {})));
                    const fixture = {
                        model: currentTask.model,
                        seed: seed,
                        relations: currentTask.relations,
                    };
                    /**
                     * Save the id to the pool by figuring out which fields are parents and which are children.
                     */
                    const poolWithFixture = insertFixtureIntoPool(fixture, newPool, orders);
                    /* Recurse */
                    const recursed = yield iterate(newTasks, poolWithFixture, iteration + 1);
                    return {
                        fixtures: [fixture, ...recursed.fixtures],
                        pool: recursed.pool,
                    };
                });
            }
            /**
             * Generates mock data from the provided model. Scalars return a mock scalar or
             * list of mock scalars, relations return an ID or lists of IDs.
             *
             * Generates mock data for scalars and relations, but skips id fields with default setting.
             */
            function getMockForTask(availablePool, otherTasks, task) {
                const initialMock = {
                    pool: availablePool,
                    tasks: otherTasks,
                    data: {},
                    include: {},
                };
                return task.model.fields
                    .filter(field => 
                /* ID fields with default shouldn't have a generated id. */
                !(field.isId && field.default !== undefined))
                    .reduce(getMockDataForField, initialMock);
                function getMockDataForField({ pool, tasks, data, include }, field) {
                    var _a;
                    const fieldModel = task.model;
                    /* Custom field mocks */
                    if ((_a = seedModels[task.model.name]) === null || _a === void 0 ? void 0 : _a.factory) {
                        const mock = seedModels[task.model.name].factory[field.name];
                        switch (typeof mock) {
                            case 'function': {
                                /* Custom function */
                                if (field.isList) {
                                    const values = mock.call(faker);
                                    return {
                                        pool,
                                        tasks,
                                        data: Object.assign(Object.assign({}, data), { [field.name]: { set: values } }),
                                        include,
                                    };
                                }
                                else {
                                    const value = mock.call(faker);
                                    return {
                                        pool,
                                        tasks,
                                        data: Object.assign(Object.assign({}, data), { [field.name]: value }),
                                        include,
                                    };
                                }
                            }
                            case 'object': {
                                /* Relation or scalar list */
                                if (field.isList) {
                                    /* Scalar list */
                                    const values = mock;
                                    return {
                                        pool,
                                        tasks,
                                        data: Object.assign(Object.assign({}, data), { [field.name]: { set: values } }),
                                        include,
                                    };
                                }
                                else {
                                    /* Relation */
                                    break;
                                }
                            }
                            case 'bigint':
                            case 'boolean':
                            case 'number':
                            case 'string': {
                                /* A constant value. */
                                const value = mock;
                                return {
                                    pool,
                                    tasks,
                                    data: Object.assign(Object.assign({}, data), { [field.name]: value }),
                                    include,
                                };
                            }
                            case 'symbol':
                            case 'undefined': {
                                /* Skipped definitions */
                                break;
                            }
                            default: {
                                throw new Error(`Unsupported type of mock "${typeof mock}".`);
                            }
                        }
                    }
                    /* ID field */
                    if (field.isId || fieldModel.idFields.includes(field.name)) {
                        switch (field.type) {
                            case scalars_1.Scalar.string: {
                                /* GUID id for strings */
                                return {
                                    pool,
                                    tasks,
                                    data: Object.assign(Object.assign({}, data), { [field.name]: faker.guid() }),
                                    include,
                                };
                            }
                            case scalars_1.Scalar.int: {
                                /* Autoincrement based on task order. */
                                return {
                                    pool,
                                    tasks,
                                    data: Object.assign(Object.assign({}, data), { [field.name]: task.order }),
                                    include,
                                };
                            }
                            default: {
                                throw new Error(`Unsupported ID type "${field.type}"`);
                            }
                        }
                    }
                    /* Scalar and relation field mocks */
                    switch (field.kind) {
                        case 'scalar': {
                            /**
                             * List scalar mocks.
                             *
                             * We provide good default mocks for functions. Anything more complex
                             * or flexible can be achieved using the exposed Chance.js library.
                             */
                            if (field.isList) {
                                switch (field.type) {
                                    /**
                                     * Scalars
                                     */
                                    case scalars_1.Scalar.string: {
                                        const strings = faker.n(faker.string, 3);
                                        return {
                                            pool,
                                            tasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: {
                                                    set: strings,
                                                } }),
                                            include,
                                        };
                                    }
                                    case scalars_1.Scalar.int: {
                                        const numbers = faker.n(faker.integer, 3, {
                                            min: -2000,
                                            max: 2000,
                                        });
                                        return {
                                            pool,
                                            tasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: { set: numbers } }),
                                            include,
                                        };
                                    }
                                    case scalars_1.Scalar.float: {
                                        const floats = faker.n(faker.floating, 3, {
                                            min: -1000,
                                            max: 1000,
                                            fixed: 2,
                                        });
                                        return {
                                            pool,
                                            tasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: { set: floats } }),
                                            include,
                                        };
                                    }
                                    case scalars_1.Scalar.date: {
                                        const dates = faker.n(faker.date, 3);
                                        return {
                                            pool,
                                            tasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: { set: dates } }),
                                            include,
                                        };
                                    }
                                    case scalars_1.Scalar.bool: {
                                        const booleans = faker.n(faker.bool, 3);
                                        return {
                                            pool,
                                            tasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: { set: booleans } }),
                                            include,
                                        };
                                    }
                                    /* Unsupported scalar */
                                    default: {
                                        throw new Error(`Unsupported scalar field ${task.model.name}${field.name} of type ${field.type}`);
                                    }
                                }
                            }
                            /**
                             * Scalar mocks.
                             */
                            switch (field.type) {
                                /**
                                 * Scalars
                                 */
                                case scalars_1.Scalar.string: {
                                    const string = faker.word();
                                    return {
                                        pool,
                                        tasks,
                                        data: Object.assign(Object.assign({}, data), { [field.name]: string }),
                                        include,
                                    };
                                }
                                case scalars_1.Scalar.int: {
                                    const number = faker.integer({
                                        min: -2000,
                                        max: 2000,
                                    });
                                    return {
                                        pool,
                                        tasks,
                                        data: Object.assign(Object.assign({}, data), { [field.name]: number }),
                                        include,
                                    };
                                }
                                case scalars_1.Scalar.float: {
                                    const float = faker.floating({
                                        min: -1000,
                                        max: 1000,
                                        fixed: 2,
                                    });
                                    return {
                                        pool,
                                        tasks,
                                        data: Object.assign(Object.assign({}, data), { [field.name]: float }),
                                        include,
                                    };
                                }
                                case scalars_1.Scalar.date: {
                                    const date = faker.date();
                                    return {
                                        pool,
                                        tasks,
                                        data: Object.assign(Object.assign({}, data), { [field.name]: date }),
                                        include,
                                    };
                                }
                                case scalars_1.Scalar.bool: {
                                    const boolean = faker.bool();
                                    return {
                                        pool,
                                        tasks,
                                        data: Object.assign(Object.assign({}, data), { [field.name]: boolean }),
                                        include,
                                    };
                                }
                                /* Unsupported scalar */
                                default: {
                                    throw new Error(`Unsupported scalar field ${task.model.name}${field.name} of type ${field.type}`);
                                }
                            }
                        }
                        /**
                         * Relations
                         *
                         * NOTE: this function assumes that we've sorted orders in
                         *   such an order that this is the most crucial step that
                         *   needs to be finished.
                         */
                        case 'object': {
                            /* Resources calculation */
                            const relation = task.relations[field.name];
                            switch (relation.type) {
                                case '1-to-1': {
                                    if (relation.field.isRequired) {
                                        /**
                                         * model A {
                                         *  b: B
                                         * }
                                         * model B {
                                         *  a: A
                                         * }
                                         *
                                         * We are creating model A.
                                         *
                                         * NOTE: field a in model B could also be optional.
                                         *      This function assumes that the order is such that
                                         *      this step should also create an instance of model B
                                         *      regardless of the meaning.
                                         */
                                        /* Creates the instances while creating itself. */
                                        /**
                                         * The other part of this relation will be taken out by the
                                         * creation step below using getMockOfModel.
                                         */
                                        const { tasks: newTasks, pool: newPool, data: instance, include: mockInclude, } = getMockOfModel(tasks, pool, relation.field);
                                        return {
                                            pool: newPool,
                                            tasks: newTasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: {
                                                    create: instance,
                                                } }),
                                            include: Object.assign(Object.assign({}, include), { [field.name]: Object.keys(mockInclude).length !== 0
                                                    ? { include: mockInclude }
                                                    : true }),
                                        };
                                    }
                                    else if (
                                    /* If this model is second made */
                                    relation.direction.from !== fieldModel.name &&
                                        !relation.backRelationField.isRequired) {
                                        /**
                                         * model A {
                                         *  b: B
                                         * }
                                         * model B {
                                         *  a?: A
                                         * }
                                         *
                                         * We are creating model B.
                                         */
                                        const units = faker.integer({
                                            min: relation.min,
                                            max: relation.max,
                                        });
                                        /* Create an instance and connect it to the relation. */
                                        const [newPool, ids] = getIDInstancesFromPool(pool, fieldModel.name, field.type, units);
                                        /**
                                         * This makes sure that relations are properly connected.
                                         */
                                        switch (ids.length) {
                                            case 0: {
                                                return {
                                                    pool: newPool,
                                                    tasks,
                                                    data,
                                                    include,
                                                };
                                            }
                                            case 1: {
                                                const [id] = ids;
                                                return {
                                                    pool: newPool,
                                                    tasks,
                                                    data: Object.assign(Object.assign({}, data), { [field.name]: {
                                                            connect: id,
                                                        } }),
                                                    include: Object.assign(Object.assign({}, include), { [field.name]: true }),
                                                };
                                            }
                                            /* istanbul ignore next */
                                            default: {
                                                throw new Error(`Something truly unexpected happened.`);
                                            }
                                        }
                                    }
                                    else {
                                        /**
                                         * It is possible that this is the other side of the relation
                                         * that has already been created in another process.
                                         *
                                         * We should't do anything in that case.
                                         */
                                        return {
                                            pool,
                                            tasks,
                                            data,
                                            include,
                                        };
                                    }
                                }
                                case '1-to-many': {
                                    if (
                                    /* This model is created first. */
                                    relation.direction.from === fieldModel.name) {
                                        /**
                                         * model A {
                                         *  bs: B[]
                                         * }
                                         * model B {
                                         *  a: A
                                         * }
                                         *
                                         * We are creating model A.
                                         */
                                        /* Skip creating the relation as we'll connect to it when creating model B. */
                                        return {
                                            pool,
                                            tasks,
                                            data,
                                            include,
                                        };
                                    }
                                    else {
                                        /* This is the second model in relation. */
                                        /**
                                         * model A {
                                         *  bs: B[]
                                         * }
                                         * model B {
                                         *  a: A
                                         * }
                                         *
                                         * We are creating model B, As already exist.
                                         */
                                        const units = faker.integer({
                                            min: relation.min,
                                            max: relation.max,
                                        });
                                        /* Create this instance and connect to others. */
                                        const [newPool, ids] = getIDInstancesFromPool(pool, fieldModel.name, field.type, units);
                                        return {
                                            pool: newPool,
                                            tasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: {
                                                    connect: ids,
                                                } }),
                                            include: Object.assign(Object.assign({}, include), { [field.name]: true }),
                                        };
                                    }
                                }
                                case 'many-to-1': {
                                    if (
                                    /* This is the first model to be created in the relation. */
                                    relation.direction.from === fieldModel.name &&
                                        relation.field.isRequired) {
                                        /**
                                         * model A {
                                         *  b: B
                                         * }
                                         * model B {
                                         *  a: A[]
                                         * }
                                         *
                                         * We are creating A.
                                         */
                                        const { tasks: newTasks, pool: newPool, data: instance, include: mockInclude, } = getMockOfModel(tasks, pool, relation.field);
                                        return {
                                            pool: newPool,
                                            tasks: newTasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: {
                                                    create: instance,
                                                } }),
                                            include: Object.assign(Object.assign({}, include), { [field.name]: Object.keys(mockInclude).length !== 0
                                                    ? { include: mockInclude }
                                                    : true }),
                                        };
                                    } /* This model is created second. */
                                    else if (relation.direction.from !== fieldModel.name) {
                                        /**
                                         * model A {
                                         *  b: B
                                         * }
                                         * model B {
                                         *  a: A[]
                                         * }
                                         *
                                         * We are creating A, B already exists.
                                         */
                                        const units = faker.integer({
                                            min: relation.min,
                                            max: relation.max,
                                        });
                                        /* Create this instance and connects it. */
                                        const [newPool, [id]] = getIDInstancesFromPool(pool, fieldModel.name, field.type, units);
                                        if (!id && relation.field.isRequired) {
                                            throw new Error(`Missing data for required relation: ${relation.field.relationName}`);
                                        }
                                        if (!id && !relation.field.isRequired) {
                                            return {
                                                pool: newPool,
                                                tasks,
                                                data,
                                                include,
                                            };
                                        }
                                        return {
                                            pool: newPool,
                                            tasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: {
                                                    connect: id,
                                                } }),
                                            include: Object.assign(Object.assign({}, include), { [field.name]: true }),
                                        };
                                    }
                                    else {
                                        /* Some other task will create this relation. */
                                        return {
                                            pool,
                                            tasks,
                                            data,
                                            include,
                                        };
                                    }
                                }
                                case 'many-to-many': {
                                    if (relation.direction.from === fieldModel.name) {
                                        /**
                                         * model A {
                                         *  bs: B[]
                                         * }
                                         * model B {
                                         *  as: A[]
                                         * }
                                         */
                                        /* Insert IDs of this instance to the pool. */
                                        return {
                                            pool,
                                            tasks,
                                            data,
                                            include,
                                        };
                                    }
                                    else {
                                        /**
                                         * model A {
                                         *  bs: B[]
                                         * }
                                         * model B {
                                         *  as: A[]
                                         * }
                                         */
                                        const units = faker.integer({
                                            min: relation.min,
                                            max: relation.max,
                                        });
                                        /* Create instances and connect to relation instances. */
                                        const [newPool, ids] = getIDInstancesFromPool(pool, fieldModel.name, field.type, units);
                                        return {
                                            pool: newPool,
                                            tasks,
                                            data: Object.assign(Object.assign({}, data), { [field.name]: {
                                                    connect: ids,
                                                } }),
                                            include: Object.assign(Object.assign({}, include), { [field.name]: true }),
                                        };
                                    }
                                }
                                /* end of relation kind switches */
                            }
                        }
                        /**
                         * Enums
                         */
                        case 'enum': {
                            const { values } = task.enums[field.name];
                            const enume = faker.pickone(values);
                            return {
                                pool,
                                tasks,
                                data: Object.assign(Object.assign({}, data), { [field.name]: enume }),
                                include,
                            };
                        }
                        /**
                         * Default field type fallback.
                         */
                        default: {
                            /* istanbul ignore next */
                            throw new Error(
                            /* prettier-ignore */
                            multilines_1.default `
                  | Unsupported field type "${field.type}".
                  | Please use a custom mock function or change your model definition.
                  `);
                        }
                    }
                }
            }
            /**
             * Recursively retrives n unique ids and removes them from the pool.
             */
            function getIDInstancesFromPool(pool, parent, child, n) {
                /* All available ids for this field (includes duplicates). */
                const allIds = lodash_1.default.get(pool, [parent, child], []);
                /* Used ids */
                let ids = [];
                let remainingIds = [];
                for (let index = 0; index < allIds.length; index++) {
                    const id = allIds[index];
                    /* Makes sure that ids are unique */
                    if (ids.some(_id => isId(_id, id))) {
                        remainingIds.push(id);
                        continue;
                    }
                    /* Fill the list. */
                    if (ids.length < n) {
                        ids.push(id);
                    }
                    else {
                        remainingIds.push(id);
                    }
                }
                if (ids.length < n) {
                    throw new Error(`Requesting more ${parent}.${child} ids than available.`);
                }
                /* Clear ids from the pool. */
                const poolWithoutIds = lodash_1.default.set(pool, [parent, child], remainingIds);
                return [poolWithoutIds, ids];
            }
            /**
             * Tells whether two ids are the same.
             *
             * @param id
             * @param comparable
             */
            function isId(id, comparable) {
                return [...Object.keys(id), ...Object.keys(comparable)].every(key => id[key] === comparable[key]);
            }
            /**
             * Inserts n-replications of ID into the pool and returns the new pool.
             */
            function insertIDInstancesIntoPool(pool, parent, child, id, n) {
                const ids = lodash_1.default.get(pool, [parent, child], []);
                return lodash_1.default.set(pool, [parent, child], [...ids, ...Array(n).fill(id)]);
            }
            /**
             * Inserts an id of a task into all parent fields it covers.
             *
             * @param id
             * @param task
             * @param pool
             */
            function insertFixtureIntoPool(fixture, initialPool, orders) {
                /**
                 * Calculates the id of this fixture.
                 */
                const id = fixture.model.fields
                    .filter(field => fixture.model.idFields.includes(field.name) || field.isId)
                    .reduce((acc, field) => {
                    if (!fixture.seed.hasOwnProperty(field.name)) {
                        throw new Error(`Return data of ${fixture.model.name} is missing id field data for ${field.name}.`);
                    }
                    return Object.assign(Object.assign({}, acc), { [field.name]: fixture.seed[field.name] });
                }, {});
                return fixture.model.fields
                    .filter(field => field.kind === 'object')
                    .reduce(insertFieldIntoPool, initialPool);
                /**
                 * This complements getMockDataForField's relations part.
                 *  * For every relation that returns unchanged mock, this should
                 *    insert id into the pool.
                 *  * For every nested create relation it should extract child fixture.
                 *
                 * @param pool
                 * @param field
                 */
                function insertFieldIntoPool(pool, field) {
                    const fieldModel = fixture.model;
                    switch (field.kind) {
                        /**
                         * Scalars, Enums
                         */
                        case 'scalar':
                        case 'enum': {
                            return pool;
                        }
                        /**
                         * Relations
                         */
                        case 'object': {
                            /* Resources calculation */
                            const relation = fixture.relations[field.name];
                            switch (relation.type) {
                                case '1-to-1': {
                                    if (relation.field.isRequired) {
                                        /**
                                         * Extracts a subfixture from compound create relation.
                                         */
                                        const subfixtureOrder = getFieldOrder(field);
                                        const subfixture = {
                                            seed: fixture.seed[field.name],
                                            model: Object.assign(Object.assign({}, subfixtureOrder.model), { fields: subfixtureOrder.model.fields.filter(({ relationName }) => relationName !== field.relationName) }),
                                            relations: utils_1.filterKeys(subfixtureOrder.relations, (key, relation) => relation.field.relationName !== field.relationName),
                                        };
                                        return insertFixtureIntoPool(subfixture, pool, orders);
                                    }
                                    else if (relation.direction.from === fieldModel.name &&
                                        !relation.field.isRequired) {
                                        /**
                                         * model A {
                                         *  b: B
                                         * }
                                         * model B {
                                         *  a: A
                                         * }
                                         *
                                         * 1-to-1 relation should take at most one id from the resource pool
                                         * and submit no new ids. Because we already manage constraints during
                                         * order creation step, we can ignore it now.
                                         */
                                        /* Insert the ID of an instance into the pool. */
                                        const newPool = insertIDInstancesIntoPool(pool, field.type, fieldModel.name, id, 1);
                                        return newPool;
                                    }
                                    else {
                                        /* Relation doesn't create any child properties. */
                                        return pool;
                                    }
                                }
                                case '1-to-many': {
                                    /**
                                     * model A {
                                     *  bs: B[]
                                     * }
                                     * model B {
                                     *  a: A
                                     * }
                                     *
                                     * 1-to-many creates the model instance and makes its ID available for later connections.
                                     */
                                    if (relation.direction.from === fieldModel.name) {
                                        /* Create the relation while creating this model instance. */
                                        // const units = faker.integer({
                                        //   min: relation.min,
                                        //   max: relation.max,
                                        // })
                                        const units = Math.max(1, relation.max);
                                        const newPool = insertIDInstancesIntoPool(pool, field.type, fieldModel.name, id, units);
                                        return newPool;
                                    }
                                    else {
                                        return pool;
                                    }
                                }
                                case 'many-to-1': {
                                    if (
                                    /* This is the first model to be created in the relation. */
                                    relation.direction.from === fieldModel.name &&
                                        relation.field.isRequired) {
                                        /**
                                         * Extracts a subfixture from compound create relation.
                                         */
                                        const subfixtureOrder = getFieldOrder(field);
                                        const subfixture = {
                                            seed: fixture.seed[field.name],
                                            model: Object.assign(Object.assign({}, subfixtureOrder.model), { fields: subfixtureOrder.model.fields.filter(({ relationName }) => relationName !== field.relationName) }),
                                            relations: utils_1.filterKeys(subfixtureOrder.relations, (key, relation) => relation.field.relationName !== field.relationName),
                                        };
                                        return insertFixtureIntoPool(subfixture, pool, orders);
                                    }
                                    else if (relation.direction.from === fieldModel.name &&
                                        !relation.field.isRequired) {
                                        /**
                                         * model A {
                                         *  b: B
                                         * }
                                         * model B {
                                         *  a: A[]
                                         * }
                                         *
                                         * Many-to-1 relations either create an instance and make it available for later use,
                                         * or connect to existing instance.
                                         */
                                        /* Insert IDs of model instance into the pool. */
                                        // const units = faker.integer({
                                        //   min: relation.min,
                                        //   max: relation.max,
                                        // })
                                        const units = Math.max(1, relation.max);
                                        const newPool = insertIDInstancesIntoPool(pool, field.type, fieldModel.name, id, units);
                                        return newPool;
                                    }
                                    else {
                                        return pool;
                                    }
                                }
                                case 'many-to-many': {
                                    /**
                                     * model A {
                                     *  bs: B[]
                                     * }
                                     * model B {
                                     *  as: A[]
                                     * }
                                     *
                                     * Many-to-many relationships simply have to follow consistency. They can either
                                     * create ID instances in the pool or connect to them.
                                     */
                                    if (relation.direction.from === fieldModel.name) {
                                        /* Insert IDs of this instance to the pool. */
                                        // const units = faker.integer({
                                        //   min: relation.min,
                                        //   max: relation.max,
                                        // })
                                        const units = Math.max(1, relation.max);
                                        const newPool = insertIDInstancesIntoPool(pool, field.type, fieldModel.name, id, units);
                                        return newPool;
                                    }
                                    else {
                                        return pool;
                                    }
                                }
                                /* end of relation kind switches */
                            }
                        }
                    }
                }
                /**
                 * Finds the order of type of the field.
                 * @param field
                 */
                function getFieldOrder(field) {
                    return orders.find(order => order.model.name === field.type);
                }
            }
            /**
             * Creates instances of a requested relation and drains the remaining tasks.
             */
            function getMockOfModel(tasks, availablePool, field) {
                // TODO: this doesn't put an id into fixture in the end.
                /* Find the requested tasks. */
                const mockTask = tasks.find(task => task.model.name === field.type);
                /* Validation check, though it should never trigger */
                /* istanbul ignore next */
                if (mockTask === undefined) {
                    throw new Error('Something very unexpected occured.');
                }
                /* Delete backrelation */
                const mockTaskWithoutBackrelation = Object.assign(Object.assign({}, mockTask), { model: Object.assign(Object.assign({}, mockTask.model), { fields: mockTask.model.fields.filter(({ relationName }) => relationName !== field.relationName) }), relations: utils_1.filterKeys(mockTask.relations, (key, relation) => relation.field.relationName !== field.relationName) });
                const remainingTasks = tasks.filter(task => task.order !== mockTask.order);
                return getMockForTask(availablePool, remainingTasks, mockTaskWithoutBackrelation);
            }
        });
    }
    /**
     * Cleverly groups the data so it's useful for the end user.
     *
     * @param fixtures
     */
    function groupFixtures(fixtures) {
        return fixtures.reduce((acc, fixture) => {
            const model = fixture.model.name;
            if (acc.hasOwnProperty(model)) {
                return Object.assign(Object.assign({}, acc), { [model]: acc[model].concat(fixture.seed) });
            }
            else {
                return Object.assign(Object.assign({}, acc), { [model]: [fixture.seed] });
            }
        }, {});
    }
}
exports.getSeed = getSeed;
//# sourceMappingURL=seed.js.map